#ifndef __SYSCALL_H__
#define __SYSCALL_H__

/* VXeR[}N */

#define SYSCALL_0(syscall_number, result)                                         \
    asm volatile("movl $%c1, %%ebx \n"                                            \
                 "int  $0x40       \n"                                            \
                 "movl %%eax, %0   \n"                                            \
                 :"=m"(result)                                                    \
                 :"g"(syscall_number)                                             \
                 :"ebx"                                                           \
                 );                                                               \

#define SYSCALL_1(syscall_number, result, arg1)                                   \
    asm volatile("movl $%c1, %%ebx \n"                                            \
                 "movl %2  , %%esi \n"                                            \
                 "int  $0x40       \n"                                            \
                 "movl %%eax, %0   \n"                                            \
                 :"=m"(result)                                                    \
                 :"g"(syscall_number), "m"(arg1)                                  \
                 :"ebx", "esi"                                                    \
                 );                                                               \

#define SYSCALL_2(syscall_number, result, arg1, arg2)                             \
    asm volatile("movl $%c1, %%ebx \n"                                            \
                 "movl %2  , %%esi \n"                                            \
                 "movl %3  , %%ecx \n"                                            \
                 "int  $0x40       \n"                                            \
                 "movl %%eax, %0   \n"                                            \
                 :"=m"(result)                                                    \
                 :"g"(syscall_number), "m"(arg1), "m"(arg2)                       \
                 :"ebx", "esi", "ecx"                                             \
                 );                                                               \

#define SYSCALL_3(syscall_number, result, arg1, arg2, arg3)                       \
    asm volatile("movl $%c1, %%ebx \n"                                            \
                 "movl %2  , %%esi \n"                                            \
                 "movl %3  , %%ecx \n"                                            \
                 "movl %4  , %%edi \n"                                            \
                 "int  $0x40       \n"                                            \
                 "movl %%eax, %0   \n"                                            \
                 :"=m"(result)                                                    \
                 :"g"(syscall_number), "m"(arg1), "m"(arg2), "m"(arg3)            \
                 :"ebx", "esi", "ecx", "edi"                                      \
                 );                                                               \

#define SYSCALL_4(syscall_number, result, arg1, arg2, arg3, arg4)                 \
    asm volatile("movl $%c1, %%ebx \n"                                            \
                 "movl %2  , %%esi \n"                                            \
                 "movl %3  , %%ecx \n"                                            \
                 "movl %4  , %%edi \n"                                            \
                 "movl %5  , %%edx \n"                                            \
                 "int  $0x40       \n"                                            \
                 "movl %%eax, %0   \n"                                            \
                 :"=m"(result)                                                    \
                 :"g"(syscall_number), "m"(arg1), "m"(arg2), "m"(arg3), "m"(arg4) \
                 :"ebx", "esi", "ecx", "edi", "edx"                               \
                 );                                                               \

#define SYSCALL_5(syscall_number, result, arg1, arg2, arg3, arg4, arg5)           \
    asm volatile("movl $%c1, %%ebx \n"                                            \
                 "movl %2  , %%esi \n"                                            \
                 "movl %3  , %%ecx \n"                                            \
                 "movl %4  , %%edi \n"                                            \
                 "movl %5  , %%edx \n"                                            \
                 "movl %6  , %%ebp \n"                                            \
                 "int  $0x40       \n"                                            \
                 "movl %%eax, %0   \n"                                            \
                 :"=m"(result)                                                    \
                 :"g"(syscall_number), "m"(arg1), "m"(arg2), "m"(arg3), "m"(arg4), "m"(arg5) \
                 :"ebx", "esi", "ecx", "edi", "edx", "ebp"                        \
                 );                                                               \

#define SYSCALL_6(syscall_number, result, arg1, arg2, arg3, arg4, arg5, arg6)     \
    asm volatile("movl $%c1, %%ebx \n"                                            \
                 "movl %2  , %%esi \n"                                            \
                 "movl %3  , %%ecx \n"                                            \
                 "movl %4  , %%edi \n"                                            \
                 "movl %5  , %%edx \n"                                            \
                 "movl %6  , %%ebp \n"                                            \
                 "movl %7  , %%esp \n"                                            \
                 "int  $0x40       \n"                                            \
                 "movl %%eax, %0   \n"                                            \
                 :"=m"(result)                                                    \
                 :"g"(syscall_number), "m"(arg1), "m"(arg2), "m"(arg3), "m"(arg4), "m"(arg5), "m"(arg6) \
                 :"ebx", "esi", "ecx", "edi", "edx", "ebp", "esp"                 \
                 );                                                               \

/* 萔錾 */

#define NULL		0
#define INT_MAX		(+0x7fffffff)
#define UINT_MAX	(+0xffffffff)
#define ULONG_MAX	UINT_MAX

#ifndef __SIZE_TYPE__
#define __SIZE_TYPE__ unsigned long
#endif
typedef __SIZE_TYPE__ size_t;

typedef unsigned char  byte;
typedef unsigned short word;
typedef unsigned int   dword;

/* VXeR[ԍ */

#define API_PUTCHAR    0x01
#define API_PUTSTR0    0x02
#define API_PUTSTR1    0x03
#define API_END        0x04
#define API_OPENWIN    0x05
#define API_PUTSTRWIN  0x06
#define API_BOXFILWIN  0x07
#define API_INITMALLOC 0x08
#define API_MALLOC     0x09
#define API_FREE       0x0A
#define API_POINT      0x0B
#define API_REFRESHWIN 0x0C
#define API_LINEWIN    0x0D
#define API_CLOSEWIN   0x0E
#define API_GETKEY     0x0F
#define API_ALLOCTIMER 0x10
#define API_INITTIMER  0x11
#define API_SETTIMER   0x12
#define API_FREETIMER  0x13
#define API_BEEP       0x14
#define API_FOPEN      0x15
#define API_FCLOSE     0x16
#define API_FSEEK      0x17
#define API_FSIZE      0x18
#define API_FREAD      0x19
#define API_CMDLINE    0x1A
#define API_GETLANG    0x1B

#define TOMO_GETMOUSE  0x1000
#define TOMO_GETTICK   0x1001
#define TOMO_RDMSR     0x1002
#define TOMO_SETLANG   0x1003
#define TOMO_SYSINFO   0x1004
#define TOMO_SYSTIME   0x1005
#define TOMO_WRMSR     0x1006

/* }EX */
typedef struct MOUSE_INFO {
	int x, y, button, state;
};

/* VXe */
typedef struct SYS_INFO {
	byte cyls;     /* u[gZN^͂ǂ܂ŃfBXNǂ񂾂̂ */
	byte leds;     /* u[g̃L[{[hLED̏ */
	byte bpps;     /* rfI[h  rbgJ[ */
	byte reserve;  /* \ */
	word screen_x; /* ʉ𑜓x */
	word screen_y; /* ʉ𑜓x */
	byte *vram;    /* VRAM̃AhX */
	word os_type;  /* OS̎ */
};

/* t */
typedef struct TIME_INFO {
	int year, month, day, hour, minutes, second;
};

/* ͂ڂOS݊VXeR[ */
extern "C" {
	void  api_putchar(char c);
	void  api_putstr0(char* s);
	void  api_putstr1(char* s, int len);
	void  api_end();
	int   api_openwin(byte* buf, int xsiz, int ysiz, int col_inv, char* title);
	void  api_putstrwin(int win, int x, int y, int col, int len, char* str);
	void  api_boxfilwin(int win, int x0, int y0, int x1, int y1, int col);
	void  api_initmalloc(void);
	dword api_malloc(int size);
	void  api_free(void* addr);
	void  api_point(int win, int x, int y, int col);
	void  api_refreshwin(int win, int x0, int y0, int x1, int y1);
	void  api_linewin(int win, int x0, int y0, int x1, int y1, int col);
	void  api_closewin(int win);
	int   api_getkey(int mode);
	int   api_alloctimer(void);
	void  api_inittimer(int timer, int data);
	void  api_settimer(int timer, int time);
	void  api_freetimer(int timer);
	void  api_beep(int tone);
	int   api_fopen(char* fname);
	void  api_fclose(int fhandle);
	void  api_fseek(int fhandle, int offset, int mode);
	int   api_fsize(int fhandle, int mode);
	int   api_fread(char* buf, int maxsize, int fhandle);
	int   api_cmdline(char *buf, int maxsize);
	int   api_getlang(void);
}

/* ͂ڂėF̉ʃVXeR[ */
extern "C" {
	void tomo_getmouse(int win, MOUSE_INFO *info);
	int  tomo_gettick(void);
	void tomo_rdmsr(int id_ecx, int *eax, int *edx);
	void tomo_setlang(int lang);
	void tomo_sysinfo(SYS_INFO *info);
	void tomo_systime(TIME_INFO *info);
	void tomo_wrmsr(int id_ecx, int eax, int edx);
	void tomo_cpuid(int id_eax, int id_ecx, int *eax, int *ebx, int *ecx, int *edx);
	void tomo_rdtsc(int *high, int *low);
}

/* ƎVXeR[ */
extern "C" {
	void HariMain(void);
	void* memcpy(void* s1, const void* s2, size_t size);
	void* memset (void *d, int c, size_t sz);
	int printf(const char* format, ...);
}

/* C++pZq */
void* operator new(size_t size);
void* operator new[](size_t size);
void  operator delete(void* address);
void  operator delete[](void* address);

#endif
